/*
 * Copyright (c) 2010 RWTH Aachen University
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include "tun_dev.h"
#include "tap-udptunnel.h"
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

int main (int argc, char *argv[]) {
	
	if (argc<5) {
		printf("usage: udp-taptunnel <destination> <local-port> <dest-port> <flowid> [<tapdevicename>]\n");
		return -1;
	}
	
	char dest[strlen(argv[1])+1] ;
	memcpy(&dest,argv[1],strlen(argv[1]));
  dest[strlen(argv[1])] = '\0';
	
	int localport = atoi(argv[2]);
	int destport = atoi(argv[3]);
	
	int flowid = htons(atoi(argv[4]));
	
	
	
	
	char devname[30];

	if (argc>5) {
		strcpy(devname,argv[5]);		
	} else {
	
	      	sprintf(devname,"tunnel"); 			
	}	

	
	//sprintf(devname,"tunnel%d",tapindex);
		
	//strcpy (devname,"tap"+itoa(tapindex));
	
	printf("Initializing...\n");
	printf("	Opening TAP Device: %s\n",devname);
	
	int tap_descriptor = tap_open((char*)&devname);
	if (tap_descriptor!=-1) {
		
		printf("	TAP device name: %s\n",(char*)&devname);
	} else {
		printf("Error! %i\n",tap_descriptor);
		return tap_descriptor;	
	}
	printf("	Binding socket!\n");
	
		
	struct sockaddr_in sin;
	sin.sin_family = AF_INET;
	sin.sin_port = htons(localport);
	sin.sin_addr.s_addr = INADDR_ANY;
	
	int sock = socket(AF_INET, SOCK_DGRAM, 0);
	
	if (bind(sock, (struct sockaddr *)&sin, sizeof(sin))==-1) {
		
		printf("Bind failed! \n");
		return -1;
	}
	
	int val_on=1;
	int sock_brdcast=setsockopt(sock, SOL_SOCKET, SO_BROADCAST, &val_on, sizeof(val_on));
	if (sock_brdcast==0) {
		printf("Successfully enabled Broadcasting over socket\n");
	} else {
		printf("Datagramm Broadcasting could not be enabled! Error code: %i\n",sock_brdcast);
	}	
	
	char buffer[3000];
	fd_set fds;
	 
	printf("Starting tunnel on port %i to %s\n",localport,dest);
	 
	 
	while (1) {
		//printf("Zeroing Filedescriptor set\n");
		FD_ZERO(&fds);
        
        FD_SET(tap_descriptor, &fds);
        FD_SET(sock, &fds);	
		
		//printf("Socket_Select\n");
		int fm = (tap_descriptor > sock ? tap_descriptor : sock) + 1;
		int f_select = select(fm, &fds, NULL, NULL, NULL);

			
		
		if(FD_ISSET(tap_descriptor, &fds) ) {
	   		//printf("TAP DEVICE is ready\n");
	   		int l = tap_read(tap_descriptor,buffer,sizeof(buffer));
       	    //printf("Read %i bytes from TAP\n",l);
       	    int offset=0; //debug! this allows to strip packet headers.
       	    //not needed! setting this to something not equal to zero messes up the ethernet frames!
       	    struct TunPacket *tpacket=(struct TunPacket*) malloc(sizeof(struct TunPacket)+(l-offset));
       		tpacket->len = htons(l-offset);
       		tpacket->flowid = flowid;
       		memcpy(&(tpacket->data),&buffer+offset,l-offset);
       		
       		struct sockaddr_in destination;

			destination.sin_family = AF_INET;
			destination.sin_addr.s_addr = inet_addr(dest);
			destination.sin_port = htons(destport);
       		
       		int bytes_sent = sendto(sock,tpacket, (l-offset)+sizeof(struct TunPacket), 0,(struct sockaddr*) &destination, sizeof(struct sockaddr_in));
			if(bytes_sent < 0) {
				printf("Error sending packet!\n");
				return (-1);
			} else {
				//printf("Sent packet, size=%i\n",bytes_sent);	
			}
			free(tpacket);
       	    
       	}
       	
       	/*
       	if(FD_ISSET(sock, &fds) ) {
       		//printf("Socket is ready\n");
	   		int l = recv(sock,buffer,sizeof(buffer),0);
	   		//printf("Read %i bytes from Socket\n",l);
           	//unwrap packet and write to socket
           	struct TunPacket *packet = (struct TunPacket*) malloc(l);
           	memcpy (packet,&buffer,l);
           	tap_write(tap_descriptor,packet->data,ntohs(packet->len));
           	//printf("Wrote %i bytes to TAP Device\n",ntohs(packet->len));
           	free(packet);
		}
  		*/
  		 	
       	if(FD_ISSET(sock, &fds) ) {
       		//printf("Socket is ready\n");
	   		int l = recv(sock,buffer,sizeof(buffer),0);
	   		//printf("Read %i bytes from Socket\n",l);
           	//unwrap packet and write to socket
           	struct TunPacket *packet = (struct TunPacket*) &buffer;
           	//memcpy (packet,&buffer,l);
           	tap_write(tap_descriptor,packet->data,ntohs(packet->len));
           	//printf("Wrote %i bytes to TAP Device\n",ntohs(packet->len));
           	//free(packet);
		}
  		
		
	}
	
	return 0;	
}
